/*
Copyright (c) 2014 Carnegie Mellon University.
All Rights Reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following acknowledgments and disclaimers.
2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following acknowledgments and disclaimers in the documentation and/or other materials provided with the distribution.
3. Products derived from this software may not include Carnegie Mellon University, "SEI and/or Software Engineering Institute" in the name of such derived product, nor shall Carnegie Mellon University, "SEI and/or Software Engineering Institute" be used to endorse or promote products derived from this software without prior written permission. For written permission, please contact permission@sei.cmu.edu.
ACKNOWLEDMENTS AND DISCLAIMERS:
Copyright 2014 Carnegie Mellon University

This material is based upon work funded and supported by the Department of Defense under Contract No. FA8721-05-C-0003 with Carnegie Mellon University for the operation of the Software Engineering Institute, a federally funded research and development center.

Any opinions, findings and conclusions or recommendations expressed in this material are those of the author(s) and do not necessarily reflect the views of the United States Department of Defense.

NO WARRANTY. THIS CARNEGIE MELLON UNIVERSITY AND SOFTWARE ENGINEERING INSTITUTE MATERIAL IS FURNISHED ON AN AS-IS BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY KIND, EITHER EXPRESSED OR IMPLIED, AS TO ANY MATTER INCLUDING, BUT NOT LIMITED TO, WARRANTY OF FITNESS FOR PURPOSE OR MERCHANTABILITY, EXCLUSIVITY, OR RESULTS OBTAINED FROM USE OF THE MATERIAL. CARNEGIE MELLON UNIVERSITY DOES NOT MAKE ANY WARRANTY OF ANY KIND WITH RESPECT TO FREEDOM FROM PATENT, TRADEMARK, OR COPYRIGHT INFRINGEMENT.

This material has been approved for public release and unlimited distribution.

Carnegie Mellon is registered in the U.S. Patent and Trademark Office by Carnegie Mellon University.

DM-0000890

 */

package edu.cmu.sei.mCOP;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Random;
import java.util.TreeSet;
import java.util.Vector;

import edu.cmu.sei.ZeroSlackBinPacker.PTask;
import edu.cmu.sei.ZeroSlackBinPacker.Packer;
import edu.cmu.sei.ZeroSlackBinPacker.Processor;
import edu.cmu.sei.ZeroSlackBinPacker.RMOrderComparator;

public class MPacker extends Packer {

	Vector<Mode> modes;
	
	Vector<ModeTransition> transitions;
	
	public Vector<ModeTransition> getModeTransitions(){
		return transitions;
	}

	Vector<MTask> mtaskset=null;
	
	public void setMTaskset(Vector<MTask> taskset){
		mtaskset = taskset;
	}
	
	public void addMTask(MTask t){
		if (mtaskset==null)
			mtaskset= new Vector<MTask>();
		mtaskset.add(t);
	}
	
	public Vector<MTask> getMTaskset(){
		return mtaskset;
	}
	
	HashMap<Mode,Vector<MTask>> globalMTasksPerMode = new HashMap<Mode,Vector<MTask>>();
	
	public Vector<MTask> getGlobalMTasksetForMode(Mode m){
		Vector<MTask>v = globalMTasksPerMode.get(m);
		if (v ==null){
			v = new Vector<MTask>();
			globalMTasksPerMode.put(m,v);
			for (MTask t:getMTaskset()){
				if (t.getModalFacade(m) != null){
					v.add(t);
				}
			}
		}
		return v;
	}

	public void setModes(Vector<Mode>modes){
		this.modes = modes;
	}
	
	public Vector<Mode>getModes(){
		return modes;
	}
	
	public MPacker(){		
	}
	
	public MPacker(Vector<Mode>modes, Vector<ModeTransition> trans, Vector<MTask> taskset){
		this.modes = modes;
		this.transitions = trans;
		setMTaskset(taskset);
	}

	TreeSet<MTask> mcCoOneModeTaskset=null;	

	public Iterator<MTask> getCritCoOneModeOrderedMTasks(Mode m){
		if (mcCoTaskset == null){
			mcCoTaskset = new TreeSet<MTask>(new DecreasingOneModeOverloadedTaskComparator(m));
			mcCoTaskset.addAll(mtaskset);
		}
		return mcCoTaskset.iterator();
		
	}

	TreeSet<MTask> mcCoTaskset=null;	

	public Iterator<MTask> getCritCoOrderedMTasks(){
		if (mcCoTaskset == null){
			mcCoTaskset = new TreeSet<MTask>(new DecreasingDominantCriticalityOverloadedTaskComparator());
			mcCoTaskset.addAll(mtaskset);
		}
		return mcCoTaskset.iterator();
		
	}

	public Iterator<MProcessor> getCrtiCoOrderedMBinList(MTask task){
		//TreeSet<MProcessor> binlist = new TreeSet<MProcessor>(new DecreasingOverloadDominantModeProcessorComparator(task));
		// Try WFD
		TreeSet<MProcessor> binlist = new TreeSet<MProcessor>(new IncreasingOverloadDominantModeProcessorComparator(task));
		for (Processor p:getProcessors())
			binlist.add((MProcessor)p);
		return binlist.iterator();
	}
	
	public Iterator<MProcessor> getCrtiCOrderedMBinList(MTask task){
		//TreeSet<MProcessor> binlist = new TreeSet<MProcessor>(new DecreasingOverloadDominantModeProcessorComparator(task));
		// Try WFD
		TreeSet<MProcessor> binlist = new TreeSet<MProcessor>(new IncreasingOverloadDominantModeProcessorComparator(task));
		for (Processor p:getProcessors())
			binlist.add((MProcessor)p);
		return binlist.iterator();
	}

	public Iterator<MProcessor> getOneModeCOrderedMBinList(Mode m){
		TreeSet<MProcessor> binlist = new TreeSet<MProcessor>(new IncreasingOverloadOneModeProcessorComparator(m));
		for (Processor p:getProcessors())
			binlist.add((MProcessor)p);
		return binlist.iterator();
		
	}

//	public void plainPack(){
//		double minimumUtilization = 0.0;
//		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
//			PTask t = iter.next();
//			minimumUtilization += ((double)t.getNormalC()) / ((double)t.getPeriod());
//		}
//		
//		long minNumProcs =  (long)Math.ceil(minimumUtilization);
//
//		// add minimum number of processors
//		for (long i=0;i<minNumProcs; i++)
//			addProcessor(new Processor("", taskset));
//		
//		for (Iterator<PTask>iter = getPlainOrderedTasks();iter.hasNext();){
//			PTask t = iter.next();
//			boolean deployed = false;
//			do{
//				for (Iterator<Processor> piter = getPlainOrderedBinListForTask(t); piter.hasNext();){
//					Processor p = piter.next();
//					deployed = p.addIfSchedulable(t);
//					if (deployed){
//						System.out.println("plainPack:Deployed: t("+t.getPeriod()+") to P("+p.getUniqueID()+"): NormalFullness("+p.normalFullnessLevel+")");						
//						break;
//					}
//				}
//				if (!deployed){
//					// add new processor
//					addProcessor(new Processor("",taskset));
//				}
//			} while (!deployed);
//		}
//	}

	public void vectorPack(int maxProcessors){
		double []minimumUtilization = new double[modes.size()];
		double maxMinUtilization;
		int numProcessors=0;
		
		for (int i=0;i<minimumUtilization.length;i++)
			minimumUtilization[i]=0;
		
		for (Iterator<MTask>iter = getCritCoOrderedMTasks();iter.hasNext();){
			MTask t = iter.next();
			int i=0;
			for (Mode m:modes){
				minimumUtilization[i++] += t.getNormalUtilization(m);
			}
		}
		maxMinUtilization = 0;
		for (int i=0;i<minimumUtilization.length;i++)
			if (maxMinUtilization<minimumUtilization[i])
				maxMinUtilization = minimumUtilization[i];
		
		long minNumProcs =  (long)Math.ceil(maxMinUtilization);

		// add minimum number of processors
		for (long i=0;i<minNumProcs && numProcessors<maxProcessors; i++){
			addProcessor(new MProcessor("",getTaskSet(), getMTaskset(), modes));
			numProcessors++;
		}
		
		// TODO:continue with modification following dominant mode order after testing 
		// schedulability in MProcessor
		// Stage one
		TreeSet<MTask> rest = new TreeSet<MTask>(new DecreasingOneModeOverloadedTaskComparator(modes.firstElement()));
		//TreeSet<MTask> rest = new TreeSet<MTask>(new OverloadedTaskComparator());
		//TreeSet<MTask> rest = new TreeSet<MTask>(new CriticalityOverloadComparator());
		for (Iterator<MTask>iter = getCritCoOneModeOrderedMTasks(modes.firstElement());iter.hasNext();){
			MTask t = iter.next();
			boolean deployed = false;
			for (Iterator<MProcessor> piter = getOneModeCOrderedMBinList(modes.firstElement()); piter.hasNext();){
				MProcessor p = piter.next();
				deployed = p.addIfSchedulable(t);
				if (deployed)
					break;
			}
			if (!deployed){
				rest.add(t);
			}
		}

		// add the rest of tasks
		for (Iterator<MTask>iter = rest.iterator();iter.hasNext();){
			MTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<MProcessor> piter = getOneModeCOrderedMBinList(modes.lastElement()); piter.hasNext();){
					MProcessor p = piter.next();
					// see if we start packing "early"
					//p.setCompression(true);
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					if (numProcessors < maxProcessors){
						MProcessor dmp =new MProcessor("",getTaskSet(), getMTaskset(), getModes());
						//dmp.setCompression(true);
						addProcessor(dmp);
						numProcessors++;
					} else {
						break;
					}
				}
			} while (!deployed);
		}
		
		// ensure that the ZSRM is run on all processors
		// so we will have an accurate Crm/Cc for each process
		for (Processor p:getProcessors()){
			MProcessor dp = (MProcessor) p;
			//dp.setCompression(true);
			dp.setCompression(false);
			dp.isSchedulable();
		}
		
	}
	
	public void copVectorPack(int maxProcessors){
		double []minimumUtilization = new double[modes.size()];
		double []maximumUtilization = new double[modes.size()];
		double maxMinUtilization;
		double maxUtilization;
		int numProcessors=0;
		
		for (int i=0;i<minimumUtilization.length;i++){
			minimumUtilization[i]=0.0;
			maximumUtilization[i]=0.0;
		}
		
		for (Iterator<MTask>iter = getCritCoOrderedMTasks();iter.hasNext();){
			MTask t = iter.next();
			int i=0;
			for (Mode m:modes){
				minimumUtilization[i] += t.getNormalUtilization(m);
				maximumUtilization[i++] += t.getOverloadedUtilization(m);
			}
		}
		maxMinUtilization = 0;
		maxUtilization = 0;
		for (int i=0;i<minimumUtilization.length;i++){
			if (maxMinUtilization<minimumUtilization[i]){
				maxMinUtilization = minimumUtilization[i];
			}
			if (maxUtilization < maximumUtilization[i]){
				maxUtilization = maximumUtilization[i];
			}
		}
		
		long minNumProcs =  (long)Math.ceil(maxMinUtilization);
		maxUtilization = maxUtilization / 0.69; // RM utilization bound
		long maxNumProcs = (long)Math.ceil(maxUtilization);
		
//		System.out.println("Min Processors: "+ minNumProcs+": Max Processors: "+maxNumProcs);

		// add minimum number of processors
		for (long i=0;i<minNumProcs && numProcessors < maxProcessors; i++){
			//addProcessor(new DualModeProcessor("",getTaskSet()));
			addProcessor(new MProcessor("",getTaskSet(), getMTaskset(), modes));
			numProcessors++;
		}
		
		// TODO:continue with modification following dominant mode order after testing 
		// schedulability in MProcessor
		// Stage one
		TreeSet<MTask> rest = new TreeSet<MTask>(new DecreasingDominantCriticalityOverloadedTaskComparator());
		//TreeSet<MTask> rest = new TreeSet<MTask>(new CriticalityOverloadComparator());
		for (Iterator<MTask>iter = getCritCoOrderedMTasks();iter.hasNext();){
			MTask t = iter.next();
			boolean deployed = false;
			for (Iterator<MProcessor> piter = getCrtiCoOrderedMBinList(t); piter.hasNext();){
				MProcessor p = piter.next();
				p.setCompression(true);
				deployed = p.addIfSchedulable(t);
				if (deployed)
					break;
			}
			if (!deployed){
				rest.add(t);
			}
		}

		// add the rest of tasks
		for (Iterator<MTask>iter = rest.iterator();iter.hasNext();){
			MTask t = iter.next();
			boolean deployed = false;
			do{
				for (Iterator<MProcessor> piter = getCrtiCOrderedMBinList(t); piter.hasNext();){
					MProcessor p = piter.next();
					// see if we start packing "early"
					//p.setCompression(true);
					deployed = p.addIfSchedulable(t);
					if (deployed)
						break;
				}
				if (!deployed){
					// add new processor
					if (numProcessors < maxProcessors){
						MProcessor dmp =new MProcessor("",getTaskSet(), getMTaskset(), getModes());
						dmp.setCompression(true);
						addProcessor(dmp);
						numProcessors++;
					} else {
						break;
					}
				}
			} while (!deployed);
		}
		
		// ensure that the ZSRM is run on all processors
		// so we will have an accurate Crm/Cc for each process
		for (Processor p:getProcessors()){
			MProcessor dp = (MProcessor) p;
			//dp.setCompression(true);
			dp.setCompression(false);
			dp.isSchedulable();
		}
	}
	
	public double getCriticalityResilience(int numCriticalityLevels){
		double resilience=0.0f;
		
		for (Mode m:modes){
			double mres = 0.0f;
			for(long ov=(1L<<(numCriticalityLevels-1))-1; ov>=0; ov>>=1){
				for(Processor p:getProcessors()){
					MProcessor mp = (MProcessor)p;
					mres += mp.getModeResiliency(ov,numCriticalityLevels,m);
				}
				if(ov == 0)
					break;
			}
			resilience += mres;
			System.out.println("Resilience of mode: "+m.uniqueID+":"+mres);
		}
		return (resilience/modes.size());
	}

	HashMap<Mode,long[]> taskPerCritByMode = new HashMap<Mode,long[]>();
	
	public long[] getNumTaskPerCriticalityForMode(Mode m, long numCriticalityLevels){
		long[] l=taskPerCritByMode.get(m);
		if (l==null){
			l=new long[(int)numCriticalityLevels];
			taskPerCritByMode.put(m,l);
			for (MTask t:getGlobalMTasksetForMode(m)){
				if(t.getModalFacade(m).criticality >= numCriticalityLevels){
					System.err.println("ERROR IN ASSUMPTIONS - CRITICALITY LEVEL ("+t.getModalFacade(m).criticality+") IS GREATER THAN NUMCRITICALITYLEVELS");
					return l;
				}
				l[(int)t.getModalFacade(m).criticality] ++;
			}			
		}
		return l;
	}
	
	public static void reassignRMPriorities(Vector<MTask> tset){
		TreeSet<MTask> reordered = new TreeSet<MTask>(new RMOrderComparator());
		reordered.addAll(tset);
		int i=0;
		for (MTask t:reordered){
			t.setPriority(i++);
		}
	}

	public static Vector<MTask> generateTaskSet(MTask []taskTypes, int numTasks){
		Vector<MTask> tset = new Vector<MTask>();
		int typeIndex=0;
		
		for (int i=0;i<numTasks;i++){
			MTask t = new MTask("t"+Integer.toString(i),
								taskTypes[typeIndex].getPeriod(),
								taskTypes[typeIndex].getOverloadedC(),
								taskTypes[typeIndex].getNormalC(),
								taskTypes[typeIndex].getModeCriticalities(),
								taskTypes[typeIndex].RmPriority,
								taskTypes[typeIndex].numModes);
			tset.add(t);
			//typeIndex = (typeIndex +1) % taskTypes.length;
			if (i != 0 && (i%(numTasks/taskTypes.length) == 0)){
				typeIndex++;
			}
		}
		
		reassignRMPriorities(tset);
		return tset;
	}


	public static Vector<MTask> generateRandomMTaskSet (int minPeriod, 
			int maxPeriod, 
			double minOverC,
			double maxOverC,
			int numTasks, 
			int utilization,			
			Vector<Mode> modes,
			int [][]criticalityArrays){
		Vector<MTask> tset = new Vector<MTask>();
		Random rand = new Random();
		
		long basePeriod = minPeriod + rand.nextInt(maxPeriod - minPeriod);
		long [] periodTypes = new long[5];
		for (int i=0; i<periodTypes.length; i++){
			int power = rand.nextInt(2);
			long period = ((long)Math.pow(2,power)) * basePeriod;
			// previous period becomes new base period
			basePeriod = period;
			periodTypes[i] = period;
		}
		double currUtilization =0.0;
		int critArrayLen = criticalityArrays.length;
		
		for (int i=0;i<numTasks;i++){
//			int power = rand.nextInt(2);
//			long period = ((long)Math.pow(2,power)) * basePeriod;
//			// previous period becomes new base period
//			basePeriod = period;
			long period = periodTypes[i%periodTypes.length];
			double overC = minOverC + (rand.nextDouble() * (maxOverC-minOverC));
			long Co = (long) (overC * (double)(period));
			long C = (long)(0.05 * (double)(period));
			//long C = (long)(rand.nextDouble()*Co);
			// C varying between 0.05 and 0.5 * Co
			//C = C + (long)(0.45 * rand.nextDouble()*Co);
			long Cc = (long)(rand.nextDouble()*Co);
			C = (Cc<C) ? C:Cc;
			HashMap<Mode,Long> modeToCriticalities = new HashMap<Mode,Long>();
			int j=0;
			for (Mode m:modes){
				modeToCriticalities.put(m,(long)criticalityArrays[i%critArrayLen][j++]);
			}
			MTask task = new MTask("t1", period, Co, C ,modeToCriticalities,0,modes.size());	
			currUtilization += task.getOverloadedUtilization();
			if (currUtilization > utilization)
				break;
			tset.add(task);
		}
		
		return tset;
	}

	public static void runFixedExperiment(){
		Vector<MTask> tset = new Vector<MTask>();
		Mode modeStreet = new Mode();
		Mode modeHighway = new Mode();

		long period = 16;
		long C = 6;
		long Co = 9;
		HashMap<Mode,Long> modeToCriticalities = new HashMap<Mode,Long>();
		modeToCriticalities.put(modeStreet, 2l); //low criticality
		modeToCriticalities.put(modeHighway, 0l);// high criticality
		MTask taskSuspension = new MTask("Suspension", period, Co, C ,modeToCriticalities,
											0, // RM Priority  -- lowest
											2);	// Num Modes
		tset.add(taskSuspension);

		modeToCriticalities = new HashMap<Mode, Long>();
		modeToCriticalities.put(modeStreet, 0l);//high
		modeToCriticalities.put(modeHighway,2l); //low
		period = 4;
		C = 2;
		Co=2;
		MTask taskPedestrian = new MTask("Pedestrian", period, Co, C, modeToCriticalities, 
										2, // RM Priority -- Highest
										2); // Num Modes
		tset.add(taskPedestrian);
		
		modeToCriticalities = new HashMap<Mode, Long>();
		modeToCriticalities.put(modeHighway, 1l); // Medium
		period=8;
		C=1;
		Co=2;
		MTask taskCruise = new MTask("Cruise", period, Co, C, modeToCriticalities,
									1,// RM Priority -- Medium
									1); // Num Modes
		
		tset.add(taskCruise);
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(modeStreet);
		modes.add(modeHighway);

		Vector <ModeTransition> transitions = new Vector<ModeTransition>();

		ModeTransition transition =new ModeTransition(modeStreet,modeHighway); 
		transitions.add(transition);
		
		MPacker packer = new MPacker(modes,transitions,tset);
		packer.copVectorPack(1);
		
		System.out.println("Suspension ZS("+taskSuspension.getModalFacade(modeStreet).ZeroSlackInstant+")");
		System.out.println("Pedestrian ZS("+taskPedestrian.getModalFacade(modeStreet).ZeroSlackInstant+")");
		//System.out.println("Cruise ZS("+taskCruise.getModalFacade(modeStreet).ZeroSlackInstant+")");

		TransitionZSRMScheduler tscheduler = new TransitionZSRMScheduler();
		tscheduler.addTask(taskSuspension);
		tscheduler.addTask(taskPedestrian);
		tscheduler.addTask(taskCruise);
		
		
		tscheduler.TransitionCompressSchedule(transition.getSourceMode(), transition.getTargetMode());
			
		System.out.println("Suspension ZS("+taskSuspension.getModalFacade(modeHighway).ZeroSlackInstant+") Transitional ZS("+taskSuspension.transZeroSlackInstant+")");
		System.out.println("Pedestrian ZS("+taskPedestrian.getModalFacade(modeHighway).ZeroSlackInstant+") Transitional ZS("+taskPedestrian.transZeroSlackInstant+")");
		System.out.println("Cruise ZS("+taskCruise.ZeroSlackInstant+")");
	}
	
	public static void runExperiment(){
		
		final int [][] [] criticalityArrays = {
				{{2,1,0}, {0,1,2}, {2,0,1}}
				};
//				{{0,1,2}, {0,1,2}, {0,1,2}},
//				{{0,1,2}, {2,1,0}, {0,2,1}},
//				{{0,2,1}, {0,2,1}, {0,2,1}},
//				{{1,0,2}, {1,0,2}, {1,0,2}},
//				{{1,2,0}, {1,2,0}, {1,2,0}},
//				{{2,0,1}, {2,0,1}, {2,0,1}},
//				{{2,1,0}, {2,1,0}, {2,1,0}},
//				{{2,1,0}, {0,1,2}, {2,0,1}}
//		};

//		final int [][] [] criticalityArrays = {
//				{{2,1,0}, {2,1,0}, {2,1,0}},
//				{{2,1,0}, {0,1,2}, {0,1,2}}
//		};
		
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		final Mode m3 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		modes.add(m3);

		Vector <ModeTransition> transitions = new Vector<ModeTransition>();
		
		transitions.add(new ModeTransition(m1,m2));
		transitions.add(new ModeTransition(m2,m3));
		transitions.add(new ModeTransition(m3,m1));


		int numExperiments = 1;//20;

		for (int j=0;j<numExperiments;j++){
			for (int i=0;i<criticalityArrays.length;i++){
				MTask [] taskTypes = new MTask[3];
				// task types
				// need to add different criticalities for different modes
				final long crit00 = (long)criticalityArrays[i][0][0];
				final long crit01 = (long)criticalityArrays[i][0][1];
				final long crit02 = (long)criticalityArrays[i][0][2];
				final long crit10 = (long)criticalityArrays[i][1][0];
				final long crit11 = (long)criticalityArrays[i][1][1];
				final long crit12 = (long)criticalityArrays[i][1][2];
				final long crit20 = (long)criticalityArrays[i][2][0];
				final long crit21 = (long)criticalityArrays[i][2][1];
				final long crit22 = (long)criticalityArrays[i][2][2];
				taskTypes[0] = new MTask("t1", 1000, 200, 50,new HashMap<Mode,Long> (){{put(m1,crit00);put(m2,crit01);put(m3,crit02);}},0,3);
				taskTypes[1] = new MTask("t2", 2000, 400, 100,new HashMap<Mode,Long> (){{put(m1,crit10);put(m2,crit11);put(m3,crit12);}},0,3);
				taskTypes[2] = new MTask("t3", 4000, 800, 200,new HashMap<Mode,Long> (){{put(m1,crit20);put(m2,crit21);put(m3,crit22);}},0,3);
				Vector<MTask> tset = generateTaskSet(taskTypes,(9*8));
				MPacker packer = new MPacker(modes,transitions,tset);
				//packer.copVectorPack(100);
				packer.copVectorPack(100);
				System.out.println("Packing [{"+crit00+","+crit01+","+crit02+"},{"+crit10+","+crit11+","+crit12+"},{"+crit20+","+crit21+","+crit22+"}]");
//				int pCnt=0;
//				for (Processor p:packer.getProcessors()){
//					System.out.println("Proc["+pCnt+++"]");
//					for (PTask t:p.getTaskSet()){
//						System.out.println("\t Task["+t.name+"]");
//					}
//				}
				System.out.println("Num Procs: "+packer.getProcessors().size());
				System.out.println("Resilience: "+packer.getCriticalityResilience(3));
				
				long [] tasksPerCrit = new long[]{0,0,0};
				for (MTask t:tset){
					for (Mode m:modes){
						PTask p = t.getModalFacade(m);
						if (p!=null){
							tasksPerCrit[(int)p.criticality]++;
						}
					}
				}
				int pCnt=0;
				for (Processor p:packer.getProcessors()){
					System.out.println("Proc["+pCnt+++"]");
//					for (PTask t:p.getTaskSet()){
//						System.out.println("\t Task["+t.name+"]");
//					}
					System.out.println("Transitional Ductility Loss: "+getTransitionalDuctilityLossForProcessor(p, transitions, modes,tasksPerCrit));
				}
								
				System.out.println("---------------------------\n");
			}
		}		
	}

	public static void runReducingProcessorsExperiment(){
//		{{0,2,2}, {2,0,0}, {2,0,0}}

		final int [][] [] criticalityArrays = {
				{{2,1,0}, {0,1,2}, {2,0,1}}
		};
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		final Mode m3 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		modes.add(m3);

		Vector <ModeTransition> transitions = new Vector<ModeTransition>();
		
		transitions.add(new ModeTransition(m1,m2));
		transitions.add(new ModeTransition(m2,m3));
		transitions.add(new ModeTransition(m3,m1));


		int numExperiments = 15;//20;
		int maxProcessors =21;

		for (int j=0;j<numExperiments;j++){
			for (int i=0;i<criticalityArrays.length;i++){
				MTask [] taskTypes = new MTask[3];
				// task types
				// need to add different criticalities for different modes
				final long crit00 = (long)criticalityArrays[i][0][0];
				final long crit01 = (long)criticalityArrays[i][0][1];
				final long crit02 = (long)criticalityArrays[i][0][2];
				final long crit10 = (long)criticalityArrays[i][1][0];
				final long crit11 = (long)criticalityArrays[i][1][1];
				final long crit12 = (long)criticalityArrays[i][1][2];
				final long crit20 = (long)criticalityArrays[i][2][0];
				final long crit21 = (long)criticalityArrays[i][2][1];
				final long crit22 = (long)criticalityArrays[i][2][2];
				taskTypes[0] = new MTask("t1", 1000, 200, 50,new HashMap<Mode,Long> (){{put(m1,crit00);put(m2,crit01);put(m3,crit02);}},0,3);
				taskTypes[1] = new MTask("t2", 2000, 400, 100,new HashMap<Mode,Long> (){{put(m1,crit10);put(m2,crit11);put(m3,crit12);}},0,3);
				taskTypes[2] = new MTask("t3", 4000, 800, 200,new HashMap<Mode,Long> (){{put(m1,crit20);put(m2,crit21);put(m3,crit22);}},0,3);
				Vector<MTask> tset = generateTaskSet(taskTypes,(9*8));
				
				MPacker packer = new MPacker(modes,transitions,tset);
				//packer.vectorPack(maxProcessors--);
				packer.copVectorPack(maxProcessors--);
				System.out.println("Packing [{"+crit00+","+crit01+","+crit02+"},{"+crit10+","+crit11+","+crit12+"},{"+crit20+","+crit21+","+crit22+"}]");

				long [] tasksPerCrit = new long[]{0,0,0};
				for (MTask t:tset){
					for (Mode m:modes){
						PTask p = t.getModalFacade(m);
						if (p!=null){
							tasksPerCrit[(int)p.criticality]++;
						}
					}
				}

				System.out.println("Num Procs: "+packer.getProcessors().size());
				System.out.println("Resilience: "+packer.getCriticalityResilience(3));
				System.out.println("---------------------------\n");

				int pCnt=0;
				for (Processor p:packer.getProcessors()){
					System.out.println("Proc["+pCnt+++"]");
					for (PTask t:p.getTaskSet()){
						//System.out.println("\t Task["+t.name+"]");
						System.out.println("Transitional Ductility: "+getTransitionalDuctilityLossForProcessor( p, transitions, modes,tasksPerCrit));
					}
				}

			}
		}		
		
	}
	
	public static void runReducingProcessorsExperimentAverages(){
//		{{0,2,2}, {2,0,0}, {2,0,0}}

		final int [][] [] criticalityArrays = {
				{{2,1,0}, {0,1,2}, {2,0,1}}
		};
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		final Mode m3 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		modes.add(m3);

		Vector <ModeTransition> transitions = new Vector<ModeTransition>();
		
		transitions.add(new ModeTransition(m1,m2));
		transitions.add(new ModeTransition(m2,m3));
		transitions.add(new ModeTransition(m3,m1));


		int numExperiments = 100;
		int numReductions = 17;//20;
		int maxProcessors =21;
		int fixMaxProcessors = maxProcessors;

		double [] vpDuctility = new double[numReductions];
		double [] copDuctility = new double[numReductions];

		for (int i=0;i<vpDuctility.length;i++){
			vpDuctility[i]=0.0;
			copDuctility[i]=0.0;
		}

		for (int k=0;k<numExperiments;k++){
			maxProcessors = fixMaxProcessors;
			for (int j=0;j<numReductions;j++){
				Vector<MTask> [] tasksets = new Vector[criticalityArrays.length];
				for (int i=0;i<criticalityArrays.length;i++){
					tasksets[i] = generateRandomMTaskSet (100, 
							1000, 
							0.1,
							0.3,
							200, 
							10,
							modes,
							criticalityArrays[i]);

				}

//				System.out.println("*****  Experiment "+j+" ******");
				for (int i=0;i<criticalityArrays.length;i++){
					Vector<MTask> tset = tasksets[i];
					Vector<MTask> tsetclone = new Vector<MTask>();
					for (MTask t:tset){
						tsetclone.add(MTask.cloneTask(t));
					}

					MPacker vppacker = new MPacker(modes,transitions,tset);
					vppacker.vectorPack(maxProcessors);
					MPacker coppacker = new MPacker(modes,transitions,tsetclone);
					coppacker.copVectorPack(maxProcessors--);

//					System.out.println("Packing [{"+
//							criticalityArrays[i][0][0]+","+
//							criticalityArrays[i][0][1]+","+
//							criticalityArrays[i][0][2]+"},{"+
//							criticalityArrays[i][1][0]+","+
//							criticalityArrays[i][1][1]+","+
//							criticalityArrays[i][1][2]+"},{"+
//							criticalityArrays[i][2][0]+","+
//							criticalityArrays[i][2][1]+","+
//							criticalityArrays[i][2][2]+"}]");

					double vpDuct = vppacker.getCriticalityResilience(3);
					vpDuctility[j] = vpDuctility[j]+vpDuct;
					double copDuct = coppacker.getCriticalityResilience(3);
					copDuctility[j] = copDuctility[j] + copDuct;
//					System.out.println("Vector Packing Num Procs: "+vppacker.getProcessors().size());
//					System.out.println("Vector Packing Resilience: "+vpDuct);
//					System.out.println("mCOP Packing Num Procs: "+coppacker.getProcessors().size());
//					System.out.println("mCOP Packing Resilience: "+copDuct);
//					System.out.println("---------------------------\n");



					// To count the number of criticality violations
					// it does not make sense. They do not vary when processors are reduced
					// does not really show the effect of the scheduler
					// commented out
					
//					long [] tasksPerCrit = new long[]{0,0,0};
//					for (MTask t:tset){
//						for (Mode m:modes){
//							PTask p = t.getModalFacade(m);
//							if (p!=null){
//								tasksPerCrit[(int)p.criticality]++;
//							}
//						}
//					}
//					int pCnt=0;
//					double critviol=0.0;
//					for (Processor p:coppacker.getProcessors()){
//						//System.out.print("Proc["+pCnt+++"]:");
//						pCnt++;
//						for (PTask t:p.getTaskSet()){
//							//System.out.println("\t Task["+t.name+"]");
//							//System.out.println("Transitional Ductility: "+getTransitionalDuctilityForProcessor( p, transitions, modes,tasksPerCrit));
//							//System.out.println("Criticality Violations: "+ getCriticalityViolationsForProcessor( p, transitions, modes,tasksPerCrit));
//							critviol+= getCriticalityViolationsForProcessor( p, transitions, modes,tasksPerCrit);
//						}
//						//System.out.println(critviol);
//					}
//					System.out.println("Criticality Violations Per Proc:"+(critviol/pCnt));
//
				}
			}		
		}
		
		for (int i=0;i<numReductions;i++){
			vpDuctility[i] = vpDuctility[i] / ((double)numExperiments);
			copDuctility[i] = copDuctility[i] / ((double)numExperiments);
			System.out.println("Paccking [{"+
					criticalityArrays[0][0][0]+","+
					criticalityArrays[0][0][1]+","+
					criticalityArrays[0][0][2]+"},{"+
					criticalityArrays[0][1][0]+","+
					criticalityArrays[0][1][1]+","+
					criticalityArrays[0][1][2]+"},{"+
					criticalityArrays[0][2][0]+","+
					criticalityArrays[0][2][1]+","+
					criticalityArrays[0][2][2]+"}]");
			System.out.println("Number of processors: "+ (fixMaxProcessors - i));
			System.out.println("Vector Packing Ductility "+vpDuctility[i]);
			System.out.println("COP Packing Ductility "+copDuctility[i]);
		}
		
	}
	
	
	public static void runBaseExperiment(){
		
		final int [][] [] criticalityArrays = {
				{{0,1,2}, {0,1,2}, {0,1,2}},
				{{0,1,2}, {2,1,0}, {0,2,1}},
				{{0,2,1}, {0,2,1}, {0,2,1}},
				{{1,0,2}, {1,0,2}, {1,0,2}},
				{{1,2,0}, {1,2,0}, {1,2,0}},
				{{2,0,1}, {2,0,1}, {2,0,1}},
				{{2,1,0}, {2,1,0}, {2,1,0}},
				{{2,1,0}, {0,1,2}, {2,0,1}}
		};

//		final int [][] [] criticalityArrays = {
//				{{2,1,0}, {2,1,0}, {2,1,0}},
//				{{2,1,0}, {0,1,2}, {0,1,2}}
//		};
		
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		final Mode m3 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		modes.add(m3);

		Vector <ModeTransition> transitions = new Vector<ModeTransition>();
		
		transitions.add(new ModeTransition(m1,m2));
		transitions.add(new ModeTransition(m2,m3));
		transitions.add(new ModeTransition(m3,m1));


		int numExperiments = 1;//20;

		for (int j=0;j<numExperiments;j++){
			for (int i=0;i<criticalityArrays.length;i++){
				MTask [] taskTypes = new MTask[3];
				// task types
				// need to add different criticalities for different modes
				final long crit00 = (long)criticalityArrays[i][0][0];
				final long crit01 = (long)criticalityArrays[i][0][1];
				final long crit02 = (long)criticalityArrays[i][0][2];
				final long crit10 = (long)criticalityArrays[i][1][0];
				final long crit11 = (long)criticalityArrays[i][1][1];
				final long crit12 = (long)criticalityArrays[i][1][2];
				final long crit20 = (long)criticalityArrays[i][2][0];
				final long crit21 = (long)criticalityArrays[i][2][1];
				final long crit22 = (long)criticalityArrays[i][2][2];
//				taskTypes[0] = new MTask("t1", 1000, 200, 50,new HashMap<Mode,Long> (){{put(m1,crit00);put(m2,crit01);put(m3,crit02);}},0);
//				taskTypes[1] = new MTask("t2", 2000, 400, 100,new HashMap<Mode,Long> (){{put(m1,crit10);put(m2,crit11);put(m3,crit12);}},0);
//				taskTypes[2] = new MTask("t3", 4000, 800, 200,new HashMap<Mode,Long> (){{put(m1,crit20);put(m2,crit21);put(m3,crit22);}},0);
//				Vector<MTask> tset = generateTaskSet(taskTypes,(9*8));
				Vector<MTask> tset = generateRandomMTaskSet (100, 
						1000, 
						0.1,
						0.3,
						200, 
						10,
						modes,
						criticalityArrays[i]);
				
//				System.out.println("taskset generated: size:"+tset.size());
				MPacker packer = new MPacker(modes,transitions,tset);
				packer.vectorPack(100);
				System.out.println("Packing [{"+crit00+","+crit01+","+crit02+"},{"+crit10+","+crit11+","+crit12+"},{"+crit20+","+crit21+","+crit22+"}]");
//				int pCnt=0;
//				for (Processor p:packer.getProcessors()){
//					System.out.println("Proc["+pCnt+++"]");
//					for (PTask t:p.getTaskSet()){
//						System.out.println("\t Task["+t.name+"]");
//					}
//				}
				System.out.println("Num Procs: "+packer.getProcessors().size());
				System.out.println("Resilience: "+packer.getCriticalityResilience(3));
				
				long [] tasksPerCrit = new long[]{0,0,0};
				for (MTask t:tset){
					for (Mode m:modes){
						PTask p = t.getModalFacade(m);
						if (p!=null){
							tasksPerCrit[(int)p.criticality]++;
						}
					}
				}
				int pCnt=0;
				for (Processor p:packer.getProcessors()){
					System.out.println("Proc["+pCnt+++"]");
//					for (PTask t:p.getTaskSet()){
//						System.out.println("\t Task["+t.name+"]");
//					}
					System.out.println("Transitional Ductility: "+getTransitionalDuctilityLossForProcessor( p, transitions, modes,tasksPerCrit));
				}

				System.out.println("---------------------------\n");
			}
		}		
	}

	public static void runComparativeExperiment(){
		
		final int [][] [] criticalityArrays = {
				{{0,1,2}, {0,1,2}, {0,1,2}},
				{{0,1,2}, {2,1,0}, {0,2,1}},
				{{0,2,1}, {0,2,1}, {0,2,1}},
				{{1,0,2}, {1,0,2}, {1,0,2}},
				{{1,2,0}, {1,2,0}, {1,2,0}},
				{{2,0,1}, {2,0,1}, {2,0,1}},
				{{2,1,0}, {2,1,0}, {2,1,0}},
				{{2,1,0}, {0,1,2}, {2,0,1}}
		};

//		final int [][] [] criticalityArrays = {
//				{{2,1,0}, {2,1,0}, {2,1,0}},
//				{{2,1,0}, {0,1,2}, {0,1,2}}
//		};
		
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		final Mode m3 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		modes.add(m3);

		Vector <ModeTransition> transitions = new Vector<ModeTransition>();
		
		transitions.add(new ModeTransition(m1,m2));
		transitions.add(new ModeTransition(m2,m3));
		transitions.add(new ModeTransition(m3,m1));

		int [] vpnumProcs = new int[criticalityArrays.length];
		int [] copnumProcs = new int[criticalityArrays.length];
		for (int i=0;i<criticalityArrays.length; i++){
			vpnumProcs[i] = 0;
			copnumProcs[i]=0;
		}

		int numExperiments = 100;

		for (int j=0;j<numExperiments;j++){
			Vector<MTask> [] tasksets = new Vector[criticalityArrays.length];
			for (int i=0;i<criticalityArrays.length;i++){
				tasksets[i] = generateRandomMTaskSet (100, 
						1000, 
						0.1,
						0.3,
						200, 
						10,
						modes,
						criticalityArrays[i]);
				
			}

			for (int i=0;i<criticalityArrays.length;i++){
				Vector<MTask> tset = tasksets[i];
				Vector<MTask> tsetclone = new Vector<MTask>();
				for (MTask t:tset){
					tsetclone.add(MTask.cloneTask(t));
				}
				
//				System.out.println("taskset generated: size:"+tset.size());
				MPacker packer = new MPacker(modes,transitions,tset);
				packer.vectorPack(100);
				vpnumProcs[i] += packer.getProcessors().size();
				MPacker copPacker = new MPacker(modes,transitions,tsetclone);
				copPacker.copVectorPack(100);
				copnumProcs[i] += copPacker.getProcessors().size();
				
//				System.out.println("Packing [{"+
//						criticalityArrays[i][0][0]+","+
//						criticalityArrays[i][0][1]+","+
//						criticalityArrays[i][0][2]+"},{"+
//						criticalityArrays[i][1][0]+","+
//						criticalityArrays[i][1][1]+","+
//						criticalityArrays[i][1][2]+"},{"+
//						criticalityArrays[i][2][0]+","+
//						criticalityArrays[i][2][1]+","+
//						criticalityArrays[i][2][2]+"}]");
//				System.out.println("Vector Packing: Num Procs: "+packer.getProcessors().size());
//				System.out.println("Vector Packing: Resilience: "+packer.getCriticalityResilience(3));
				
//				System.out.println("mCOP Packing: Num Procs: "+copPacker.getProcessors().size());
//				System.out.println("mCOP Packing: Resilience: "+copPacker.getCriticalityResilience(3));
//				
//				System.out.println("---------------------------\n");
			}
		}
		for (int i=0;i<vpnumProcs.length;i++){
			vpnumProcs[i] = vpnumProcs[i]/numExperiments;
			copnumProcs[i] = copnumProcs[i] / numExperiments;
			System.out.println("Packing [{"+
					criticalityArrays[i][0][0]+","+
					criticalityArrays[i][0][1]+","+
					criticalityArrays[i][0][2]+"},{"+
					criticalityArrays[i][1][0]+","+
					criticalityArrays[i][1][1]+","+
					criticalityArrays[i][1][2]+"},{"+
					criticalityArrays[i][2][0]+","+
					criticalityArrays[i][2][1]+","+
					criticalityArrays[i][2][2]+"}]");
			System.out.println("Vector Packing: Num Procs: "+vpnumProcs[i]);
			System.out.println("mCOP Packing: Num Procs: "+ copnumProcs[i]);
		}
	}

//	public static double getTransitionalDuctilityForProcessor(Processor p, Vector<ModeTransition>transitions, Vector<Mode>modes, long [] tasksPerCriticality){
//		double ret=0.0;
//		
//		TransitionZSRMScheduler tscheduler = new TransitionZSRMScheduler();
//		tscheduler.addAllTasks(((MProcessor)p).globalMTaskset);
//
//		for (ModeTransition tran:transitions){
//			tscheduler.TransitionCompressSchedule(tran.getSourceMode(), tran.getTargetMode());
//			ret += tscheduler.getTransitionalDuctility(tran.getSourceMode(),tran.getTargetMode(),3,tasksPerCriticality,transitions.size());
//		}
//		
//		return ret;
//	}
	
	public static double getTransitionalDuctilityLossForProcessor(Processor p, Vector<ModeTransition>transitions, Vector<Mode>modes, long [] tasksPerCriticality){
		double ret=0.0;
		
		int numCriticalities = tasksPerCriticality.length;
		double [] ductFactorPerCriticality = new double[numCriticalities];
		
		for (int i=0;i<ductFactorPerCriticality.length;i++){
			ductFactorPerCriticality[i] = Math.pow((numCriticalities+1),i);
		}
		
		TransitionZSRMScheduler tscheduler = new TransitionZSRMScheduler();
		tscheduler.addAllTasks(((MProcessor)p).globalMTaskset);

		for (ModeTransition tran:transitions){
			// transZSInstant should be initialized to zero in zsrm
			tscheduler.TransitionCompressSchedule(tran.getSourceMode(), tran.getTargetMode());
			//ret +=tscheduler.getProportionOfCriticalityViolations(tran.getSourceMode(),tran.getTargetMode(),3,tasksPerCriticality,transitions.size(),ductFactorPerCriticality);
			ret +=tscheduler.getProportionalTransitionalDuctilityLoss(tran.getSourceMode(),tran.getTargetMode(),3,tasksPerCriticality,transitions.size(),ductFactorPerCriticality);
		}		
		return ((double)ret)/((double) transitions.size());		
	}

	public static void testCannonicalExample(){
		final Mode m1 = new Mode();
		final Mode m2 = new Mode();
		final Mode m3 = new Mode();
		
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(m1);
		modes.add(m2);
		modes.add(m3);
		
		Vector <ModeTransition> transitions = new Vector<ModeTransition>();
		
		transitions.add(new ModeTransition(m1,m2));
		transitions.add(new ModeTransition(m2,m3));
		transitions.add(new ModeTransition(m3,m1));
		
		
		MTask t1 = new MTask("ts", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,0l);put(m3,0l);}},0,3);
		MTask t2 = new MTask("tb", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,0l);put(m3,2l);}},0,3);
		MTask t3 = new MTask("tt", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t4 = new MTask("ta", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t5 = new MTask("ts", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,0l);put(m3,0l);}},0,3);
		MTask t6 = new MTask("tb", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,0l);put(m3,2l);}},0,3);
		MTask t7 = new MTask("tt", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t8 = new MTask("ta", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t9 = new MTask("ts", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,0l);put(m3,0l);}},0,3);
		MTask t10 =new MTask("tb", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,0l);put(m3,2l);}},0,3);
		MTask t11 =new MTask("tt", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t12 =new MTask("ta", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t13 =new MTask("ts", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,0l);put(m3,0l);}},0,3);
		MTask t14 =new MTask("tb", 400, 100, 50,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,0l);put(m3,2l);}},0,3);
		MTask t15 =new MTask("tt", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,0l);put(m2,2l);put(m3,1l);}},1,3);
		MTask t16 =new MTask("ta", 800, 250, 100,new HashMap<Mode,Long> (){{put(m1,2l);put(m2,2l);put(m3,1l);}},1,3);

		Vector<MTask> taskset = new Vector<MTask>();
		taskset.add(t1);
		taskset.add(t2);
		taskset.add(t3);
		taskset.add(t4);
		taskset.add(t5);
		taskset.add(t6);
		taskset.add(t7);
		taskset.add(t8);

		taskset.add(t9);
		taskset.add(t10);
		taskset.add(t11);
		taskset.add(t12);
		taskset.add(t13);
		taskset.add(t14);
		taskset.add(t15);
		taskset.add(t16);

		MPacker packer = new MPacker(modes, transitions, taskset);
		
		packer.copVectorPack(100);
		//packer.vectorPack(100);
		
		long [] tasksPerCrit = new long[]{0,0,0};
		for (MTask t:taskset){
			for (Mode m:modes){
				PTask p = t.getModalFacade(m);
				if (p!=null){
					tasksPerCrit[(int)p.criticality]++;
				}
			}
		}
		
		int pCnt=0;
		for (Processor p:packer.getProcessors()){
			System.out.println("Proc["+pCnt+++"]");
			for (PTask t:p.getTaskSet()){
				System.out.println("\t Task["+t.name+"]");
			}
			System.out.println("Transitional Ductility: "+getTransitionalDuctilityLossForProcessor( p, transitions, modes,tasksPerCrit));
		}
		
		System.out.println("Resilience: "+packer.getCriticalityResilience(3));
		
	}
	
	public static void testRandomTaskGeneration(){
		Vector<Mode> modes = new Vector<Mode>();
		modes.add(new Mode());
		modes.add(new Mode());
		modes.add(new Mode());
		
		int [][] criticalityArrays = new int[][]{
				{1,2,3},
				{1,2,3}
				};
		
		Vector<MTask> tasks = generateRandomMTaskSet (100, 
				1000, 
				0.1,
				0.3,
				10, 
				10,
				modes,
				criticalityArrays);
		
		double utilization=0.0;
		for (MTask t:tasks){
			System.out.println("Task["+t.getUniqueID()+"]:");
			for (Mode m:modes){
				PTask pt = t.getModalFacade(m);
				System.out.println("\t Mode facade: T("+pt.getPeriod()+"), Co("+pt.getOverloadedC()+"), C("+pt.getNormalC()+"), Crit("+pt.criticality+")");
			}
			utilization += t.getOverloadedUtilization();
		}
		System.out.println("Utilization: "+utilization);
	}
	
	public static void main(String[] args){
//		runReducingProcessorsExperiment();
//		testCannonicalExample();
//		System.out.println("----------Base Experiment ------------");
//		runBaseExperiment();
//		
//		System.out.println("\n\n--------mCOP Experiment-----------------");
		runExperiment();
//		runComparativeExperiment();
		
		
//		runReducingProcessorsExperimentAverages();
//		runFixedExperiment();
	}
	
	
}
